unit cpu;

interface

type
  UINT = LongWord;
  DWORD = LongWord;

const
  R1_H   = $0001; R2_H   = $0010; R3_H   = $0100;
  R1_8   = $0002; R2_8   = $0020; R3_8   = $0200;
  R1_16  = $0003; R2_16  = $0030; R3_16  = $0300;
  R1_32  = $0004; R2_32  = $0040; R3_32  = $0400;
  R1_64  = $0005; R2_64  = $0050; R3_64  = $0500;
  R1_128 = $0006; R2_128 = $0060; R3_128 = $0600;
  R1_MEM = $0008; R2_MEM = $0080; R3_MEM = $0800;

  IMM2_8  = $01000020;
  IMM2_16 = $01000030;
  IMM2_32 = $01000040;
  IMM2_64 = $01000050;
  IMM3_8  = $01000200;
  IMM3_16 = $01000300;
  IMM3_32 = $01000400;
  IMM3_64 = $01000500;

  LEA_X1  = $00000000;
  LEA_X2  = $00010000;
  LEA_X4  = $00020000;
  LEA_X8  = $00030000;

  REP_ECX  = $00;
  REP_RCX  = $80;
  REP_CMPS = 1;
  REP_INS  = 2;
  REP_LODS = 3;
  REP_MOVS = 4;
  REP_OUT  = 5;
  REP_SCAS = 6;
  REP_STOS = 7;

  NONE = 0;

  AH = 1;
  BH = 2;
  CH = 3;
  DH = 4;

  AL   = 5;  AX   = 6;  EAX  = 7;  RAX = 8;
  BL   = 9;  BX   = 10; EBX  = 11; RBX = 12;
  CL   = 13; CX   = 14; ECX  = 15; RCX = 16;
  DL   = 17; DX   = 18; EDX  = 19; RDX = 20;
  SIL  = 21; SI   = 22; ESI  = 23; RSI = 24;
  DIL  = 25; DI   = 26; EDI  = 27; RDI = 28;
  BPL  = 29; BP   = 30; EBP  = 31; RBP = 32;
  SPL  = 33; SP   = 34; ESP  = 35; RSP = 36;
  R8B  = 37; R8W  = 38; R8D  = 39; R8  = 40;
  R9B  = 41; R9W  = 42; R9D  = 43; R9  = 44;
  R10B = 45; R10W = 46; R10D = 47; R10 = 48;
  R11B = 49; R11W = 50; R11D = 51; R11 = 52;
  R12B = 53; R12W = 54; R12D = 55; R12 = 56;
  R13B = 57; R13W = 58; R13D = 59; R13 = 60;
  R14B = 61; R14W = 62; R14D = 63; R14 = 64;
  R15B = 65; R15W = 66; R15D = 67; R15 = 68;
  IPL  = 69; IP   = 70; EIP  = 71; RIP = 72;

  cpu_namereg_lower : array [1..72] of WideString = (
    'ah',   'bh',   'ch',  'dh',
    'al',   'ax',   'eax',  'rax',
    'bl',   'bx',   'ebx',  'rbx',
    'cl',   'cx',   'ecx',  'rcx',
    'dl',   'dx',   'edx',  'rdx',
    'sil',  'si',   'esi',  'rsi',
    'dil',  'di',   'edi',  'rdi',
    'bpl',  'bp',   'ebp',  'rbp',
    'spl',  'sp',   'esp',  'rsp',
    'r8b',  'r8w',  'r8d',  'r8',
    'r9b',  'r9w',  'r9d',  'r9',
    'r10b', 'r10w', 'r10d', 'r10',
    'r11b', 'r11w', 'r11d', 'r11',
    'r12b', 'r12w', 'r12d', 'r12',
    'r13b', 'r13w', 'r13d', 'r13',
    'r14b', 'r14w', 'r14d', 'r14',
    'r15b', 'r15w', 'r15d', 'r15',
    'ipl',  'ip',   'eip',  'rip');

  cpu_namemem_lower : array [2..5] of WideString = (
    'byte ptr','word ptr', 'dword ptr', 'qword ptr');

  ud_cpu : array [0..68] of Int64 = (
  NONE,
  AL,   CL,   DL,   BL,
  AH,   CH,   DH,   BH,
  SPL,  BPL,  SIL,  DIL,
  R8B,  R9B,  R10B, R11B,
  R12B, R13B, R14B, R15B,
  AX,   CX,   DX,   BX,
  SP,   BP,   SI,   DI,
  R8W,  R9W,  R10W, R11W,
  R12W, R13W, R14W, R15W,
  EAX,  ECX,  EDX,  EBX,
  ESP,  EBP,  ESI,  EDI,
  R8D,  R9D,  R10D, R11D,
  R12D, R13D, R14D, R15D,
  RAX,  RCX,  RDX,  RBX,
  RSP,  RBP,  RSI,  RDI,
  R8,   R9,   R10,  R11,
  R12,  R13,  R14,  R15);

type
  PCpuReg = ^TCpuReg;
  TCpuReg = type Int64;

  TCpuFlag = (fCF,fRESERVED1_1,fPF,fRESERVED3_0,fAF,fRESERVED5_0,fZF,fSF,fTF,fIF,fDF,fOF,fIO,
    fNF,fRESERVED15_0,fRF,fVM,fAC,fVF,fVP,fID);
  TCpuFlags = set of TCpuFlag;

  TCpuJump = (jJA,jJAE,jJB,jJBE,jJC,jJE,jJG,jJGE,jJL,jJLE,jJNA,jJNAE,jJNB,jJNBE,jJNC,jJNE,jJNG,
    jJNGE,jJNL,jJNLE,jJNO,jJNP,jJNS,jJNZ,jJO,jJP,jJPE,jJPO,jJS,jJZ);
  TCpuJumps = set of TCpuJump;

  TCpuPrefix = (pADR,pLOCK,pOPR,pSTR2,pSTR3,pREX);
  TCpuPrefixes = set of TCpuPrefix;

{$L cpu.obj}
function cpu_rex(const Register:TCpuReg):HRESULT;external;
function cpu_rextest1(const R1:TCpuReg):HRESULT;external;
function cpu_rextest2(const R1,R2:TCpuReg):HRESULT;external;
function cpu_adc(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_add(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_and(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_bsf(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_bsr(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_bswap(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_bt(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_btc(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_btr(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_bts(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cbw(var RAX:Int64):HRESULT;external;
function cpu_cwde(var RAX:Int64):HRESULT;external;
function cpu_cdqe(var RAX:Int64):HRESULT;external;
function cpu_clc(var F:TCpuFlags):HRESULT;external;
function cpu_cld(var F:TCpuFlags):HRESULT;external;
function cpu_cmc(var F:TCpuFlags):HRESULT;external;
function cpu_cmovcc(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cmp(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cmps(var RSI,RDI,S,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cmpxchg(var RAX,R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cpuid(var RAX,RBX,RCX,RDX:Int64):HRESULT;external;
function cpu_crc32(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_cwd(var RAX,RDX:Int64):HRESULT;external;
function cpu_cdq(var RAX,RDX:Int64):HRESULT;external;
function cpu_cqo(var RAX,RDX:Int64):HRESULT;external;
function cpu_dec(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_div(var RAX,RDX,R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_idiv(var RAX,RDX,R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_imul(var RAX,RDX,R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_imul2(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_imul3(var R1,R2,IMM:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_inc(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_jcc(var RIP,IMM:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_jmp(var RIP,R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_lea(var R1,R2,R3:Int64;const Imm32:Int32;const P:DWORD):HRESULT;external;
function cpu_lods(var RAX,RSI,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_mov(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_movs(var RSI,RDI,S,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_movsx(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_movsxd(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_movzx(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_mul(var RAX,RDX,R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_neg(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_nop:HRESULT;external;
function cpu_not(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_or(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_pause:HRESULT;external;
function cpu_pop(var RSP,R1,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_popf(var RSP,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_popfw(var RSP,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_popfd(var RSP,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_popfq(var RSP,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_popcnt(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_push(var RSP,R1,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_pushf(var RSP,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_pushfw(var RSP,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_pushfd(var RSP,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_pushfq(var RSP,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_rcl(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_rcr(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_rol(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_ror(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_rdtsc(var RAX,RDX:Int64):HRESULT;external;
function cpu_rdtscp(var RAX,RDX:Int64):HRESULT;external;
function cpu_sal(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_sar(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_shl(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_shr(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_sbb(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_scas(var RAX,RDI,S:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_setcc(var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_shld(var R1,R2,RCX:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_shrd(var R1,R2,RCX:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_stc(var F:TCpuFlags):HRESULT;external;
function cpu_std(var F:TCpuFlags):HRESULT;external;
function cpu_stos(var RAX,RDI,D:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_sub(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_test(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_xadd(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_xchg(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;
function cpu_xor(var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;external;

function IMM8(const Imm8:Int8):TCpuReg;INLINE;
function IMM16(const Imm16:Int16):TCpuReg;INLINE;
function IMM32(const Imm32:Int32):TCpuReg;INLINE;
function IMM64(const Imm64:Int64):TCpuReg;INLINE;
function BYTE_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;INLINE;
function WORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;INLINE;
function DWORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;INLINE;
function QWORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;INLINE;

const
  { Instruction is not encodable. }
  CORE_CPU_E_NE = HRESULT($A00C0440);

type
  TCPU = class
  protected type
    TFunc1 = function (var R1:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;
    TFunc2 = function (var R1,R2:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;
    TFunc3 = function (var R1,R2,R3:Int64;var F:TCpuFlags;const P:DWORD):HRESULT;
  protected
    FReg    : array [0..16] of Int64;
    FTmp    : array [0..31] of Byte;
    FXmm    : array [0..15] of array [0..15] of Byte;
    FMode   : UINT;
    FMax    : UINT;
    FMem    : UINT;
    FFlags  : TCpuFlags;
  protected
    function Register1(const Register:Integer;out Param:DWORD):PInt64;
    function Register2(const Register:Integer;out Param:DWORD):PInt64;
    function Rex1(const Dst:TCpuReg;const Func:TFunc1):HRESULT;
    function Rex2(const Dst,Src:TCpuReg;const Func:TFunc2):HRESULT;
    function Rex3(const Dst,Src,Ex:TCpuReg;const Func:TFunc3):HRESULT;
    function RexP(const Register:TCpuReg;const M:Pointer;const Func:Pointer):HRESULT;
  public
    constructor Create;
    function GetReg(const Register:Byte):Int64;
    procedure SetReg(const Register:Byte;const Value:Int64);
    function GetRmv(const Index:Integer):Int64;
    procedure SetRmv(const Index:Integer;const Value:Int64);
    function GetMode:UINT;
    procedure SetMode(Value:UINT);
    function GetFlags:TCpuFlags;
    procedure SetFlags(const Value:TCpuFlags);
    function Memory(const Exp:TCpuReg;var Value:Int64;const Mode:Byte=0):HRESULT;
    function Eadc(const Dst,Src:TCpuReg):HRESULT;
    function Eadd(const Dst,Src:TCpuReg):HRESULT;
    function Eand(const Dst,Src:TCpuReg):HRESULT;
    function Ebsf(const Dst,Src:TCpuReg):HRESULT;
    function Ebsr(const Dst,Src:TCpuReg):HRESULT;
    function Ebswap(const Dst:TCpuReg):HRESULT;
    function Ebt(const Dst,Src:TCpuReg):HRESULT;
    function Ebtc(const Dst,Src:TCpuReg):HRESULT;
    function Ebtr(const Dst,Src:TCpuReg):HRESULT;
    function Ebts(const Dst,Src:TCpuReg):HRESULT;
    function Ecbw:HRESULT;
    function Ecwde:HRESULT;
    function Ecdqe:HRESULT;
    function Eclc:HRESULT;
    function Ecld:HRESULT;
    function Ecmc:HRESULT;
    function Ecmovcc(const Condition:TCpuJump;const Dst,Src:TCpuReg):HRESULT;
    function Ecmp(const Dst,Src:TCpuReg):HRESULT;
    //function Ecmpsb:HRESULT;
    //function Ecmpsw:HRESULT;
    //function Ecmpsd:HRESULT;
    //function Ecmpsq:HRESULT;
    function Ecmpxchg(const Dst,Src:TCpuReg):HRESULT;
    function Ecpuid:HRESULT;
    function Ecrc32(const Dst,Src:TCpuReg):HRESULT;
    function Ecwd:HRESULT;
    function Ecdq:HRESULT;
    function Ecqo:HRESULT;
    function Edec(const Dst:TCpuReg):HRESULT;
    function Ediv(const Src:TCpuReg):HRESULT;
    function Eidiv(const Src:TCpuReg):HRESULT;
    function Eimul(const Src:TCpuReg):HRESULT;overload;
    function Eimul(const Dst,Src:TCpuReg):HRESULT;overload;
    function Eimul(const Dst,Src,Imm:TCpuReg):HRESULT;overload;
    function Einc(const Dst:TCpuReg):HRESULT;
    function Ejcc(const Condition:TCpuJump;const Dst:TCpuReg):HRESULT;
    function Ejmp(const Dst:TCpuReg):HRESULT;
    function Elea(const Dst,Src:TCpuReg):HRESULT;
    //function Elodsb:HRESULT;
    //function Elodsw:HRESULT;
    //function Elodsd:HRESULT;
    //function Elodsq:HRESULT;
    function Emov(const Dst,Src:TCpuReg;const Imm64:Int64=0):HRESULT;
    //function Emovsb:HRESULT;
    //function Emovsw:HRESULT;
    //function Emovsd:HRESULT;
    //function Emovsq:HRESULT;
    function Emovsx(const Dst,Src:TCpuReg):HRESULT;
    function Emovsxd(const Dst,Src:TCpuReg):HRESULT;
    function Emovzx(const Dst,Src:TCpuReg):HRESULT;
    function Emul(const Src:TCpuReg):HRESULT;
    function Eneg(const Dst:TCpuReg):HRESULT;
    function Enop:HRESULT;
    function Enot(const Dst:TCpuReg):HRESULT;
    function Eor(const Dst,Src:TCpuReg):HRESULT;
    function Epause:HRESULT;
    function Epop(const Dst:TCpuReg):HRESULT;
    function Epopfw:HRESULT;
    function Epopfd:HRESULT;
    function Epopfq:HRESULT;
    function Epopcnt(const Dst,Src:TCpuReg):HRESULT;
    function Epush(const Src:TCpuReg):HRESULT;
    function Epushfw:HRESULT;
    function Epushfd:HRESULT;
    function Epushfq:HRESULT;
    function Ercl(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Ercr(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Erol(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Eror(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Erdtsc:HRESULT;
    function Erdtscp:HRESULT;
    function Esal(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Esar(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Eshl(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Eshr(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
    function Esbb(const Dst,Src:TCpuReg):HRESULT;
    //function Escasb:HRESULT;
    //function Escasw:HRESULT;
    //function Escasd:HRESULT;
    //function Escasq:HRESULT;
    function Esetcc(const Condition:TCpuJump;const Dst:TCpuReg):HRESULT;
    function Eshld(const Dst,Src:TCpuReg;const SrcEx:TCpuReg=CL):HRESULT;
    function Eshrd(const Dst,Src:TCpuReg;const SrcEx:TCpuReg=CL):HRESULT;
    function Estc:HRESULT;
    function Estd:HRESULT;
    //function Estosb:HRESULT;
    //function Estosw:HRESULT;
    //function Estosd:HRESULT;
    //function Estosq:HRESULT;
    function Esub(const Dst,Src:TCpuReg):HRESULT;
    function Etest(const Dst,Src:TCpuReg):HRESULT;
    function Exadd(const Dst,Src:TCpuReg):HRESULT;
    function Exchg(const Dst,Src:TCpuReg):HRESULT;
    function Exor(const Dst,Src:TCpuReg):HRESULT;
    //function Decode(const Code:PByte;out Instruction:TCPUInstruction):HRESULT;
    //function Emulate(const Instruction:TCPUInstruction):HRESULT;
    property Reg[const Register:Byte]:Int64 read GetReg write SetReg;default;
    property Rmv[const Index:Integer]:Int64 read GetRmv write SetRmv;
    property Mode:UINT read GetMode write SetMode;
    property Flags:TCpuFlags read GetFlags write SetFlags;
  end;

implementation

function IMM8(const Imm8:Int8):TCpuReg;
begin Result := Int64(Imm8) shl 32 or Int64($02) shl 24;end;

function IMM16(const Imm16:Int16):TCpuReg;
begin Result := Int64(Imm16) shl 32 or Int64($03) shl 24;end;

function IMM32(const Imm32:Int32):TCpuReg;
begin Result := Int64(Imm32) shl 32 or Int64($04) shl 24;end;

function IMM64(const Imm64:Int64):TCpuReg;
begin Result := $05000000;end;

function BYTE_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;
begin Result := Int64(Imm) shl 32 or Int64($82) shl 24 or Int64(R1) or Int64(R2) shl 8 or Mul shl 16;end;

function WORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;
begin Result := Int64(Imm) shl 32 or Int64($83) shl 24 or Int64(R1) or Int64(R2) shl 8 or Mul shl 16;end;

function DWORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;
begin Result := Int64(Imm) shl 32 or Int64($84) shl 24 or Int64(R1) or Int64(R2) shl 8 or Mul shl 16;end;

function QWORD_PTR(const R1:TCpuReg;const R2:TCpuReg=NONE;const Mul:Byte=0;const Imm:Int32=0):TCpuReg;
begin Result := Int64(Imm) shl 32 or Int64($85) shl 24 or Int64(R1) or Int64(R2) shl 8 or Mul shl 16;end;

constructor TCPU.Create;
begin
SetMode(64);
end;

function TCPU.Register1(const Register:Integer;out Param:DWORD):PInt64;
asm
  mov       r11,rcx
  mov       r10d,edx
  shr       r10d,31
  shl       r10d,3
  movzx     edx,dl
  test      edx,edx
  jz        @@exit
  cmp       edx,72
  ja        @@exit
  lea       rax,TCPU[rcx].[FReg]
  cmp       edx,4
  jbe       @@high
  mov       ecx,edx
  sub       ecx,5
  shr       ecx,2
  lea       rax,[rax+rcx*8]
  shl       ecx,2
  sub       edx,ecx
  sub       edx,3
  cmp       edx,TCPU[r11].[FMax]
  je        @@exit
  or        edx,r10d
  mov       [r8],edx
  ret

@@high:
  lea       rax,[rax+rdx*8-08]
  or        r10d,R1_H
  mov       [r8],r10d
  ret

@@exit:
  xor       rax,rax
  mov       [r8],eax
end;

function TCPU.Register2(const Register:Integer;out Param:DWORD):PInt64;
asm
  mov       r11,rcx
  mov       r10d,edx
  shr       r10d,31
  shl       r10d,7
  movzx     edx,dl
  test      edx,edx
  jz        @@exit
  cmp       edx,72
  ja        @@exit
  lea       rax,TCPU[rcx].[FReg]
  cmp       edx,4
  jbe       @@high
  mov       ecx,edx
  sub       ecx,5
  shr       ecx,2
  lea       rax,[rax+rcx*8]
  shl       ecx,2
  sub       edx,ecx
  sub       edx,3
  cmp       edx,TCPU[r11].[FMax]
  je        @@exit
  shl       edx,4
  or        edx,r10d
  mov       [r8],edx
  ret

@@high:
  lea       rax,[rax+rdx*8-08]
  or        r10d,R1_H
  mov       [r8],r10d
  ret

@@exit:
  xor       rax,rax
  mov       [r8],eax
end;

function TCPU.Rex1(const Dst:TCpuReg;const Func:TFunc1):HRESULT;
asm
  .NOFRAME
  xchg      rdx,rcx
  call      cpu_rextest1
  test      eax,$80000000
  jnz       @@ne
  cmp       TCPU[rdx].[FMode],64
  je        @@start
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne

@@start:
  push      rbx
  push      rdi
  push      r8
  push      r9
  push      0
  sub       rsp,$20
  mov       rbx,rdx
  mov       rdi,rcx
  test      edi,$80000000
  jz        @@r
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,rdi
  shr       eax,24
  and       eax,15
  or        eax,R1_MEM
  mov       [rsp+$20],eax
  jmp       @@next

@@r:
  mov       rcx,rbx
  mov       rdx,rdi
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  test      rax,rax
  jz        @@exit
  mov       rdi,rax

@@next:
  mov       rax,offset cpu_setcc
  cmp       rax,[rsp+$30]
  jne       @@call
  movzx     eax,byte ptr [rsp+$28]
  shl       eax,16
  or        [rsp+$20],eax

@@call:
  mov       rcx,rdi
  lea       rdx,TCPU[rbx].[FFlags]
  mov       r8d,[rsp+$20]
  call      qword ptr [rsp+$30]
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$38
  pop       rdi
  pop       rbx
  ret

@@exit:
  add       rsp,$38
  pop       rdi
  pop       rbx

@@ne:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Rex2(const Dst,Src:TCpuReg;const Func:TFunc2):HRESULT;
asm
  .NOFRAME
  xchg      rdx,rcx
  xchg      rdx,r8
  call      cpu_rextest2
  test      eax,$80000000
  jnz       @@ne
  cmp       TCPU[r8].[FMode],64
  je        @@start
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne
  xchg      rdx,rcx
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne
  xchg      rdx,rcx

@@start:
  push      rbx
  push      rdi
  push      rsi
  push      r10
  push      r9
  push      0
  push      0
  sub       rsp,$20
  mov       rbx,r8
  mov       rdi,rcx
  mov       rsi,rdx

@@test1:
  test      edi,$80000000
  jz        @@r1
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,rdi
  shr       eax,24
  and       eax,15
  or        eax,R1_MEM
  mov       [rsp+$20],eax
  jmp       @@test2

  .ALIGN 16
@@r1:
  mov       rcx,rbx
  mov       rdx,rdi
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  test      rax,rax
  jz        @@exit
  mov       rdi,rax
  jmp       @@test2

@@test2:
  test      esi,$80000000
  jz        @@r2
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,rsi
  shr       eax,24
  and       eax,15
  shl       eax,4
  or        eax,R2_MEM
  mov       [rsp+$24],eax
  jmp       @@next

  .ALIGN 16
@@r2:
  test      si,si
  jz        @@i
  mov       rcx,rbx
  mov       rdx,rsi
  lea       r8,[rsp+$24]
  call      TCPU.Register2
  test      rax,rax
  jz        @@exit
  mov       rsi,rax
  jmp       @@next

  .ALIGN 16
@@i:
  mov       rax,rsi
  shr       rax,32
  shr       esi,24
  and       esi,7
  shl       esi,4
  or        esi,$01000000
  mov       [rsp+$24],esi
  cmp       sil,$50
  cmove     rax,[rsp+$38]
  lea       rsi,[rsp+$28]
  mov       [rsi],rax

  .ALIGN 16
@@next:
  mov       rax,offset cpu_cmovcc
  cmp       rax,[rsp+$30]
  je        @@cmovcc
  mov       rax,offset cpu_jcc
  cmp       rax,[rsp+$30]
  je        @@jcc
  jmp       @@call

@@cmovcc:
  movzx     eax,byte ptr [rsp+$38]
  shl       eax,16
  or        [rsp+$20],eax
  jmp       @@call

@@jcc:
  movzx     ecx,byte ptr [rsp+$38]
  shl       ecx,16
  or        [rsp+$24],ecx
  jmp       @@call

  .ALIGN 16
@@call:
  mov       rcx,rdi
  mov       rdx,rsi
  lea       r8,TCPU[rbx].[FFlags]
  mov       r9d,[rsp+$20]
  or        r9d,[rsp+$24]
  call      qword ptr [rsp+$30]
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$40
  pop       rsi
  pop       rdi
  pop       rbx
  ret

@@exit:
  add       rsp,$40
  pop       rsi
  pop       rdi
  pop       rbx

@@ne:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Rex3(const Dst,Src,Ex:TCpuReg;const Func:TFunc3):HRESULT;
asm
  .NOFRAME
  xchg      rdx,rcx
  xchg      rdx,r8
  call      cpu_rextest2
  test      eax,$80000000
  jnz       @@ne
  cmp       TCPU[r8].[FMode],64
  je        @@start
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne
  xchg      rdx,rcx
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne
  xchg      rcx,r9
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne
  xchg      rcx,r9
  xchg      rcx,rdx

@@start:
  push      rbx
  push      rdi
  push      rsi
  push      r15
  push      r10
  push      r11
  push      0
  push      0
  push      0
  sub       rsp,$20
  mov       rbx,r8
  mov       rdi,rcx
  mov       rsi,rdx
  mov       r15,r9

@@test1:
  test      edi,$80000000
  jz        @@r1
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,rdi
  shr       eax,24
  and       eax,15
  or        eax,R1_MEM
  mov       [rsp+$20],eax
  jmp       @@test2

  .ALIGN 16
@@r1:
  mov       rcx,rbx
  mov       rdx,rdi
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  test      rax,rax
  jz        @@exit
  mov       rdi,rax
  jmp       @@test2

@@test2:
  test      esi,$80000000
  jz        @@r2
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,rsi
  shr       eax,24
  and       eax,15
  shl       eax,4
  or        eax,R2_MEM
  mov       [rsp+$24],eax
  jmp       @@test3

  .ALIGN 16
@@r2:
  mov       rcx,rbx
  mov       rdx,rsi
  lea       r8,[rsp+$24]
  call      TCPU.Register2
  test      rax,rax
  jz        @@exit
  mov       rsi,rax
  jmp       @@test3

@@test3:
  test      r15d,$80000000
  jnz       @@exit
  test      r15w,r15w
  jz        @@imm
  mov       rcx,rbx
  mov       rdx,r15
  lea       r8,[rsp+$28]
  call      TCPU.Register1
  test      rax,rax
  jz        @@exit
  mov       r15,rax
  shl       dword ptr [rsp+$28],8
  jmp       @@next

@@imm:
  mov       rax,r15
  shr       rax,32
  shr       r15d,24
  and       r15d,7
  shl       r15d,8
  or        r15d,$01000000
  mov       [rsp+$28],r15d
  cmp       r15w,$500
  cmove     rax,[rsp+$38]
  lea       r15,[rsp+$30]
  mov       [r15],rax

@@next:
  mov       rax,offset cpu_imul3
  cmp       rax,[rsp+$40]
  je        @@call
  movddup   xmm0,[rsp+$40]
  movdqa    xmm1,xmm0
  movdqa    xmm2,xmm0
  pcmpeqq   xmm0,dqword ptr [@@asmtable+$00]
  ptest     xmm0,xmm0
  jnz       @@call
  pcmpeqq   xmm1,dqword ptr [@@asmtable+$10]
  pcmpeqq   xmm2,dqword ptr [@@asmtable+$20]
  por       xmm1,xmm2
  ptest     xmm1,xmm1
  jz        @@exit
  mov       qword ptr [rsp+$24],0
  jmp       @@call

@@call:
  mov       rcx,rdi
  mov       rdx,rsi
  mov       r8,r15
  lea       r9,TCPU[rbx].[FFlags]
  mov       ebx,[rsp+$20]
  or        ebx,[rsp+$24]
  or        ebx,[rsp+$28]
  mov       [rsp+$20],ebx
  call      rax
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$48
  pop       r15
  pop       rsi
  pop       rdi
  pop       rbx
  ret

@@exit:
  add       rsp,$48
  pop       r15
  pop       rsi
  pop       rdi
  pop       rbx

@@ne:
  mov       eax,CORE_CPU_E_NE
  ret

  .ALIGN 16
@@asmtable: dq cpu_shld, cpu_shrd, cpu_div, cpu_idiv, cpu_mul, cpu_imul
end;

function TCPU.RexP(const Register:TCpuReg;const M:Pointer;const Func:Pointer):HRESULT;
asm
  .NOFRAME
  xchg      rdx,rcx
  mov       rax,offset cpu_push
  cmp       rax,r9
  je        @@push
  mov       rax,offset cpu_pop
  cmp       rax,r9
  jne       @@ne
  test      ecx,$80000000
  jnz       @@test
  test      cx,cx
  jnz       @@test
  mov       eax,CORE_CPU_E_NE
  ret

@@push:
  test      cx,cx
  jz        @@start

@@test:
  call      cpu_rextest1
  test      eax,$80000000
  jnz       @@ne
  cmp       TCPU[rdx].[FMode],64
  je        @@start
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne

@@start:
  push      rbx
  push      r12
  push      r13
  push      r8
  push      r9
  push      0
  push      0
  sub       rsp,$20
  mov       rbx,rdx
  mov       r12,rcx
  mov       r13,r8
  test      r12d,$80000000
  jz        @@reg
  lea       rax,TCPU[rbx].[FTmp]
  xchg      rax,r12
  shr       eax,24
  and       eax,15
  shl       eax,4
  or        eax,R2_MEM
  mov       [rsp+$24],eax
  jmp       @@next

@@reg:
  test      r12w,r12w
  jz        @@imm
  mov       rcx,rbx
  mov       rdx,r12
  lea       r8,[rsp+$24]
  call      TCPU.Register2
  test      rax,rax
  jz        @@exit
  mov       r12,rax
  jmp       @@next

@@imm:
  mov       rax,r12
  shr       rax,32
  shr       r12d,24
  and       r12d,7
  shl       r12d,4
  or        r12d,$01000000
  mov       [rsp+$24],r12d
  lea       r12,[rsp+$28]
  mov       [r12],rax

@@next:
  cmp       TCPU[rbx].[FMode],64
  sete      al
  movzx     eax,al
  add       eax,4
  or        eax,[rsp+$24]
  mov       [rsp+$20],eax
  lea       rcx,[TCPU[rbx].[FReg]+$38]
  mov       rdx,r12
  mov       r8,r13
  lea       r9,TCPU[rbx].[FFlags]
  call      qword ptr [rsp+$30]
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$40
  pop       r13
  pop       r12
  pop       rbx
  ret

@@exit:
  add       rsp,$40
  pop       r13
  pop       r12
  pop       rbx

@@ne:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.GetReg(const Register:Byte):Int64;
asm
  movzx     edx,dl
  lea       rax,TCPU[rcx].[FReg]
  test      edx,edx
  jz        @@exit
  cmp       edx,4
  jbe       @@1
  cmp       edx,72
  ja        @@exit
  mov       ecx,edx
  sub       ecx,5
  shr       ecx,2
  lea       rax,[rax+rcx*8]
  shl       ecx,2
  sub       edx,ecx
  sub       edx,3
  lea       rcx,[@@jmptable]
  jmp       [rcx+rdx*8-08]

@@1:
  movzx     rax,byte ptr [rax+rdx*8-08+1]
  ret

@@2:
  movzx     rax,byte ptr [rax]
  ret

@@3:
  movzx     rax,word ptr [rax]
  ret

@@4:
  mov       eax,[rax]
  ret

@@5:
  mov       rax,[rax]
  ret

@@exit:
  xor       rax,rax
  ret

  .ALIGN 16
@@jmptable: dq @@1, @@2, @@3, @@4, @@5
end;

procedure TCPU.SetReg(const Register:Byte;const Value:Int64);
asm
  movzx     edx,dl
  lea       rax,TCPU[rcx].[FReg]
  test      edx,edx
  jz        @@exit
  cmp       edx,4
  jbe       @@1
  cmp       edx,72
  ja        @@exit
  mov       ecx,edx
  sub       ecx,5
  shr       ecx,2
  lea       rax,[rax+rcx*8]
  shl       ecx,2
  sub       edx,ecx
  sub       edx,3
  lea       rcx,[@@jmptable]
  jmp       [rcx+rdx*8-08]

@@1:
  mov       [rax+rdx*8-08+1],r8b
  ret

@@2:
  mov       [rax],r8b
  ret

@@3:
  mov       [rax],r8w
  ret

@@4:
  and       r8d,$FFFFFFFF

@@5:
  mov       [rax],r8
  ret

@@exit:
  xor       rax,rax
  ret

  .ALIGN 16
@@jmptable: dq @@1, @@2, @@3, @@4, @@5
end;

function TCPU.GetRmv(const Index:Integer):Int64;
asm
  xor       rax,rax
  cmp       edx,1
  ja        @@leave
  mov       rax,qword ptr TCPU[rcx].[FTmp+rdx*8]

@@leave:
end;

procedure TCPU.SetRmv(const Index:Integer;const Value:Int64);
asm
  cmp       edx,1
  ja        @@leave
  mov       qword ptr TCPU[rcx].[FTmp+rdx*8],r8

@@leave:
end;

function TCPU.GetMode:UINT;
asm
  mov       eax,TCPU[rcx].[FMode]
end;

procedure TCPU.SetMode(Value:UINT);
begin
Value := (UINT(Value = 64) + 1) shl 5;
if FMode <> Value then
begin
  FMode := Value;
  case Value of
  32:
    begin
      FMax := 5;
      FMem := R1_32 or R2_32;
    end;
  64:
    begin
      FMax := 0;
      FMem := R1_64 or R2_64;
    end;
  end;
end;
end;

function TCPU.GetFlags:TCpuFlags;
asm
  mov       eax,TCPU[rcx].[FFlags]
end;

procedure TCPU.SetFlags(const Value:TCpuFlags);
asm
  mov       TCPU[rcx].[FFlags],edx
end;

function TCPU.Memory(const Exp:TCpuReg;var Value:Int64;const Mode:Byte=0):HRESULT;
asm
  test      edx,$80000000
  jz        @@ne
  test      dx,dx
  jz        @@immediate
  xchg      rdx,rcx
  call      cpu_rextest1
  test      eax,$80000000
  jnz       @@ne
  cmp       TCPU[rdx].[FMode],64
  je        @@start
  call      cpu_rex
  xor       eax,S_FALSE
  jnz       @@ne

@@start:
  xchg      rdx,rcx
  push      rbx
  push      rsi
  push      r12
  push      r13
  push      r8
  push      r9
  push      0
  sub       rsp,$20
  mov       rbx,rcx
  mov       rsi,rdx
  xor       r12,r12
  xor       r13,r13
  test      dh,dh
  jz        @@r1

@@r2:
  mov       rcx,rbx
  mov       dl,dh
  lea       r8,[rsp+$24]
  call      TCPU.Register2
  mov       r13,rax

@@r1:
  test      sil,sil
  jz        @@z1
  mov       rcx,rbx
  mov       rdx,rsi
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  mov       r12,rax

@@z1:
  test      r12,r12
  jnz       @@z2
  lea       r12,[rsp+$28]
  mov       eax,[rsp+$24]
  shr       eax,4
  or        [rsp+$24],eax

@@z2:
  test      r13,r13
  jnz       @@call
  lea       r13,[rsp+$28]
  mov       eax,[rsp+$20]
  shl       eax,4
  or        [rsp+$20],eax

@@call:
  movzx     eax,byte ptr [rsp+$28]
  mov       qword ptr [rsp+$28],0
  test      eax,eax
  jnz       @@auto
  mov       eax,TCPU[rbx].[FMode]
  shr       al,4
  add       al,1
  movzx     eax,al

@@auto:
  mov       [rsp+$20],eax
  mov       r9,rsi
  shr       r9,32                      //Imm32
  shr       esi,16
  cmp       sil,8
  sete      al
  shr       sil,1
  sub       sil,al                     //Mul
  movzx     esi,sil
  shl       esi,16
  or        esi,[rsp+$20]
  or        esi,[rsp+$24]
  and       esi,NOT (R1_MEM or R2_MEM)
  mov       rcx,[rsp+$30]
  mov       rdx,r12                    //R2
  mov       r8,r13                     //R3
  mov       [rsp+$20],esi              //Params
  call      cpu_lea
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$38
  pop       r13
  pop       r12
  pop       rsi
  pop       rbx
  ret

@@immediate:
  push      0
  push      0
  push      0
  sub       rsp,$20
  mov       r9,rdx
  shr       r9,32                      //Imm
  shr       rdx,16
  cmp       dl,8
  sete      al
  shr       dl,1
  sub       dl,al                      //Mul
  movzx     edx,dl
  shl       edx,16
  or        edx,TCPU[rcx].[FMem]
  mov       [rsp+$20],edx
  mov       rcx,r8
  lea       rdx,[rsp+$28]              //R2 = 0
  lea       r8,[rsp+$28]               //R3 = 0
  call      cpu_lea
  mov       ecx,CORE_CPU_E_NE
  test      eax,$80000000
  cmovnz    eax,ecx
  add       rsp,$38
  ret

@@ne:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Eadc(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_adc
  jmp       REX2
end;

function TCPU.Eadd(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_add
  jmp       REX2
end;

function TCPU.Eand(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_and
  jmp       REX2
end;

function TCPU.Ebsf(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_bsf
  jmp       REX2
end;

function TCPU.Ebsr(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_bsr
  jmp       REX2
end;

function TCPU.Ebswap(const Dst:TCpuReg):HRESULT;
asm
  mov       r8,offset cpu_bswap
  jmp       REX1
end;

function TCPU.Ebt(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_bt
  jmp       REX2
end;

function TCPU.Ebtc(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_btc
  jmp       REX2
end;

function TCPU.Ebtr(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_btr
  jmp       REX2
end;

function TCPU.Ebts(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_bts
  jmp       REX2
end;

function TCPU.Ecbw:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]
  jmp       cpu_cbw
end;

function TCPU.Ecwde:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]
  jmp       cpu_cwde
end;

function TCPU.Ecdqe:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]
  jmp       cpu_cdqe
end;

function TCPU.Eclc:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FFlags]
  jmp       cpu_clc
end;

function TCPU.Ecld:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FFlags]
  jmp       cpu_cld
end;

function TCPU.Ecmc:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FFlags]
  jmp       cpu_cmc
end;

function TCPU.Ecmovcc(const Condition:TCpuJump;const Dst,Src:TCpuReg):HRESULT;
asm
  movzx     r10,dl
  mov       rdx,r8
  mov       r8,r9
  mov       r9,offset cpu_cmovcc
  jmp       REX2
end;

function TCPU.Ecmp(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_cmp
  jmp       REX2
end;

function TCPU.Ecmpxchg(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_cmpxchg
  jmp       REX2
end;

function TCPU.Ecpuid:HRESULT;
asm
  lea       rax,TCPU[rcx].[FReg]
  lea       rcx,[rax+$00]              //RAX
  lea       rdx,[rax+$08]              //RBX
  lea       r8,[rax+$10]               //RCX
  lea       r9,[rax+$18]               //RDX
  jmp       cpu_cpuid
end;

function TCPU.Ecrc32(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_crc32
  jmp       REX2
end;

function TCPU.Ecwd:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]       //RAX
  lea       rdx,[rcx+$18]              //RDX
  jmp       cpu_cwd
end;

function TCPU.Ecdq:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]       //RAX
  lea       rdx,[rcx+$18]              //RDX
  jmp       cpu_cdq
end;

function TCPU.Ecqo:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]       //RAX
  lea       rdx,[rcx+$18]              //RDX
  jmp       cpu_cqo
end;

function TCPU.Edec(const Dst:TCpuReg):HRESULT;
asm
  mov       r8,offset cpu_dec
  jmp       REX1
end;

function TCPU.Ediv(const Src:TCpuReg):HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  setne     al
  mov       r9,rdx
  mov       edx,8   //RAX
  mov       r8d,20  //RDX
  sub       dl,al
  sub       r8b,al
  mov       r10,offset cpu_div
  jmp       REX3
end;

function TCPU.Eidiv(const Src:TCpuReg):HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  setne     al
  mov       r9,rdx
  mov       edx,8   //RAX
  mov       r8d,20  //RDX
  sub       dl,al
  sub       r8b,al
  mov       r10,offset cpu_idiv
  jmp       REX3
end;

function TCPU.Eimul(const Src:TCpuReg):HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  setne     al
  mov       r9,rdx
  mov       edx,8   //RAX
  mov       r8d,20  //RDX
  sub       dl,al
  sub       r8b,al
  mov       r10,offset cpu_imul
  jmp       REX3
end;

function TCPU.Eimul(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_imul2
  jmp       REX2
end;

function TCPU.Eimul(const Dst,Src,Imm:TCpuReg):HRESULT;
asm
  mov       r10,offset cpu_imul3
  jmp       REX3
end;

function TCPU.Einc(const Dst:TCpuReg):HRESULT;
asm
  mov       r8,offset cpu_inc
  jmp       REX1
end;

function TCPU.Ejcc(const Condition:TCpuJump;const Dst:TCpuReg):HRESULT;
asm
  test      r8w,r8w
  jnz       @@exit
  cmp       TCPU[rcx].[FMode],64
  setne     al
  movzx     eax,al
  movzx     r10,dl
  mov       edx,72                     //RIP
  sub       edx,eax
  mov       r9,offset cpu_jcc
  jmp       REX2

@@exit:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Ejmp(const Dst:TCpuReg):HRESULT;
asm
  test      dx,dx
  jz        @@imm
  push      rcx
  push      rdx
  push      0
  sub       rsp,$20
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  test      rax,rax
  lea       rsp,[rsp+$20]
  pop       r8
  pop       rdx
  pop       rcx
  jz        @@exit
  and       r8d,NOT (R1_MEM OR R2_MEM)
  cmp       r8d,3
  jbe       @@exit
  cmp       TCPU[rcx].[FMode],64
  sete      ah
  cmp       r8d,5
  sete      al
  cmp       al,ah
  jne       @@exit

@@imm:
  mov       r8,rdx
  mov       edx,72                     //RIP
  mov       r9,offset cpu_jmp
  jmp       REX2

@@exit:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Elea(const Dst,Src:TCpuReg):HRESULT;
asm
  push      rcx
  push      r8
  push      0
  sub       rsp,$20
  lea       r8,[rsp+$20]
  call      TCPU.Register1
  add       rsp,$20
  pop       r9
  pop       rdx
  pop       rcx
  mov       r8,rax
  test      r8,r8
  jnz       TCPU.Memory
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Emov(const Dst,Src:TCpuReg;const Imm64:Int64=0):HRESULT;
asm
  mov       r10,r9
  mov       r9,offset cpu_mov
  cmp       TCPU[rcx].[FMode],64
  je        REX2
  mov       eax,r8d
  shr       eax,24
  and       eax,7
  cmp       eax,5
  jne       REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Emovsx(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_movsx
  jmp       REX2
end;

function TCPU.Emovsxd(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_movsxd
  jmp       REX2
end;

function TCPU.Emovzx(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_movzx
  jmp       REX2
end;

function TCPU.Emul(const Src:TCpuReg):HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  setne     al
  mov       r9,rdx
  mov       edx,8   //RAX
  mov       r8d,20  //RDX
  sub       dl,al
  sub       r8b,al
  mov       r10,offset cpu_mul
  jmp       REX3
end;

function TCPU.Eneg(const Dst:TCpuReg):HRESULT;
asm
  mov       r8,offset cpu_neg
  jmp       REX1
end;

function TCPU.Enop:HRESULT;
asm
  jmp       cpu_nop
end;

function TCPU.Enot(const Dst:TCpuReg):HRESULT;
asm
  mov       r8,offset cpu_not
  jmp       REX1
end;

function TCPU.Eor(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_or
  jmp       REX2
end;

function TCPU.Epause:HRESULT;
asm
  jmp       cpu_pause
end;

function TCPU.Epop(const Dst:TCpuReg):HRESULT;
asm
  lea       r8,[TCPU[rcx].[FTmp]+$08]
  mov       r9,offset cpu_pop
  jmp       REXP
end;

function TCPU.Epopfw:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  sete      r9b
  movzx     r9d,r9b
  add       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_popfw
end;

function TCPU.Epopfd:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  je        @@exit
  mov       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_popfd

@@exit:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Epopfq:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  sete      r9b
  movzx     r9d,r9b
  add       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_popfq
end;

function TCPU.Epopcnt(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_popcnt
  jmp       REX2
end;

function TCPU.Epush(const Src:TCpuReg):HRESULT;
asm
  lea       r8,[TCPU[rcx].[FTmp]+$08]
  mov       r9,offset cpu_push
  jmp       REXP
end;

function TCPU.Epushfw:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  sete      r9b
  movzx     r9d,r9b
  add       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_pushfw
end;

function TCPU.Epushfd:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  je        @@exit
  mov       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_pushfd

@@exit:
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Epushfq:HRESULT;
asm
  cmp       TCPU[rcx].[FMode],64
  sete      r9b
  movzx     r9d,r9b
  add       r9d,4
  lea       r8,TCPU[rcx].[FFlags]
  lea       rdx,[TCPU[rcx].[FTmp]+$08]
  lea       rcx,[TCPU[rcx].[FReg]+$38]
  jmp       cpu_pushfq
end;

function TCPU.Ercl(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_rcl
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Ercr(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_rcr
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Erol(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_rol
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Eror(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_ror
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Erdtsc:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]       //RAX
  lea       rdx,[rcx+$18]              //RDX
  jmp       cpu_rdtsc
end;

function TCPU.Erdtscp:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FReg]       //RAX
  lea       rdx,[rcx+$18]              //RDX
  jmp       cpu_rdtscp
end;

function TCPU.Esal(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_sal
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Esar(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_sar
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Eshl(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_shl
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Eshr(const Dst:TCpuReg;const Src:TCpuReg=CL):HRESULT;
asm
  mov       r9,offset cpu_shr
  test      r8w,r8w
  jz        REX2
  cmp       r8d,13
  je        REX2
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Esbb(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_sbb
  jmp       REX2
end;

function TCPU.Esetcc(const Condition:TCpuJump;const Dst:TCpuReg):HRESULT;
asm
  movzx     r9d,dl
  mov       rdx,r8
  mov       r8,offset cpu_setcc
  jmp       REX1
end;

function TCPU.Eshld(const Dst,Src:TCpuReg;const SrcEx:TCpuReg=CL):HRESULT;
asm
  mov       r10,offset cpu_shld
  test      r9w,r9w
  jz        REX3
  cmp       r9d,13
  je        REX3
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Eshrd(const Dst,Src:TCpuReg;const SrcEx:TCpuReg=CL):HRESULT;
asm
  mov       r10,offset cpu_shrd
  test      r9w,r9w
  jz        REX3
  cmp       r9d,13
  je        REX3
  mov       eax,CORE_CPU_E_NE
end;

function TCPU.Estc:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FFlags]
  jmp       cpu_stc
end;

function TCPU.Estd:HRESULT;
asm
  lea       rcx,TCPU[rcx].[FFlags]
  jmp       cpu_std
end;

function TCPU.Esub(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_sub
  jmp       REX2
end;

function TCPU.Etest(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_test
  jmp       REX2
end;

function TCPU.Exadd(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_xadd
  jmp       REX2
end;

function TCPU.Exchg(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_xchg
  jmp       REX2
end;

function TCPU.Exor(const Dst,Src:TCpuReg):HRESULT;
asm
  mov       r9,offset cpu_xor
  jmp       REX2
end;


end.